import random

import pygame
from pygame.locals import *

import imagecache


num_initial_balls = 100
megabyte = 2 ** 20

# make an image cache
default_expiration = 5.0  # seconds
max_memory = 350 * megabyte
image_cache = imagecache.ImageCache(default_expiration, max_memory)

# Results of various steps (len(range(127, 255, step)) ** 3):
# 8 creates 4096 base images
# 16 creates 512 base images
# 24 creates 216 base images
# 32 creates 64 base images
# 48 creates 27 base images
# 64 creates 8 base images
step = 32
color_component_range = range(127, 255, step)
colors = []


def load_base_images():
    for r in color_component_range:
        for g in color_component_range:
            for b in color_component_range:
                color = (r, g, b)
                colors.append(color)
                image = pygame.Surface((64, 64))
                image.fill(color)
                # N.B.: Our base images must not be allowed to expire. We look
                # them up in cache when we have to make a new transformation.
                # The cache's default for entities is None: no aging.
                image_cache.add(color, imagecache.vanilla_key, image)


class Ball(object):

    def __init__(self):
        # our ident in the cache
        self._color = random.choice(colors)

        # our transforms in the cache
        rot, flipy, flipx, zoom, alpha = imagecache.vanilla_key
        self._rot = rot
        self._flip_y = flipy
        self._flip_x = flipx
        self._zoom = zoom
        self._alpha = alpha
        self._cache_key = self._make_key()

        # our blitting stuff
        self._image = None
        self.rect = None

        # init our image and rect - this is also triggered by transforms
        self._init_image()

        # set a randon location on the screen
        screen = pygame.display.get_surface()
        screen_rect = screen.get_rect()
        self.rect.x = random.randrange(screen_rect.x, screen_rect.right - self.rect.w)
        self.rect.y = random.randrange(screen_rect.y, screen_rect.bottom - self.rect.h)

        # set a random rotation rate
        self._rot_speed = (random.random() + 0.1) * random.choice((-1, 1))

    def update(self):
        """a little behavior to cause real-time interaction with the cache"""
        rot = (self._rot + self._rot_speed) % 360.0
        self.set('_rot', rot)

    @property
    def image(self):
        """if None, then image is uninitialized or a transform attr has changed"""
        if self._image is None:
            self._cache_key = self._make_key()
            self._init_image()
        return self._image

    def set(self, attr, value):
        """a hack to avoid coding a bunch of decorators to trigger the dirty state"""
        current = getattr(self, attr)
        if current != value:
            setattr(self, attr, value)
            self._image = None

    def _init_image(self):
        """called when self._image is None

        Look up in cache. If not found, then create the image by transforming base image.
        """
        image = image_cache.get(self._color, self._cache_key)
        if image is None:
            image = image_cache.get(self._color, imagecache.vanilla_key)
            # getting the base image, better not return None
            assert image is not None
            # apply the transforms to the base image
            if self._rot != 0.0 or self._zoom != 1.0:
                image = pygame.transform.rotozoom(image, self._rot, self._zoom)
                if self._rot not in (0.0, 90.0, 180.0, 270.0):
                    image.set_colorkey((0, 0, 0))
            if self._alpha != 255:
                image.set_alpha(self._alpha)
            # flipy and flipx not implemented for this sprite

            # add the image using the cache's default aging
            image_cache.add(self._color, self._cache_key, image, -1)
        self._image = image
        rect = self.rect
        if rect is None:
            rect = self._image.get_rect()
        else:
            rect = self._image.get_rect(center=rect.center)
        self.rect = rect

    def _make_key(self):
        """when a transform triggers a new image, we need to update our key

        This is for efficiency only. We could construct a new key on the fly,
        but that is wasteful.
        """
        rot = int(round(self._rot)) % 360
        zoom = round(self._zoom, 1)
        alpha = int(round(self._alpha))
        return rot, self._flip_x, self._flip_y, zoom, int(alpha)


class Game(object):

    def __init__(self):
        self.max_fps = 60
        self.dt = 1.0 / self.max_fps / 1.0
        self.fps_elapsed = self.max_fps
        self.clock = pygame.time.Clock()

        self.balls = []
        self.make_balls(num_initial_balls)

        # tally balls each tick
        self.hits = 0
        self.misses = 0

        # run-time behavior states
        self.running = False

    def run(self):
        """loop it"""
        self.running = True
        while self.running:
            self.update()
            self.draw()
            self.clock.tick(self.max_fps)

    def update(self):
        """update the game"""
        self.update_balls()
        image_cache.update()
        self.update_fps()

        for e in pygame.event.get():
            if e.type == KEYDOWN:
                if e.key == K_ESCAPE:
                    self.running = False
                elif e.key == K_a:
                    image_cache.enable_age_cap(not image_cache.age_cap_is_enabled())
                elif e.key == K_m:
                    image_cache.enable_memory_cap(not image_cache.memory_cap_is_enabled())
                elif e.key in (K_PLUS, K_EQUALS):
                    self.make_balls(25)
                elif e.key == K_MINUS:
                    if len(self.balls) > 25:
                        self.remove_balls(25)

    def update_balls(self):
        for ball in self.balls:
            ball.update()

    def update_fps(self):
        """update the caption"""
        self.fps_elapsed += self.dt
        if self.fps_elapsed >= 0.25:
            self.fps_elapsed -= 0.25
            pygame.display.set_caption(
                '{} fps | Balls {} | Age {} | Mem {} | Cached {} at {} MB | Hits {}% | Purged mem:{} old:{}'.format(
                    int(self.clock.get_fps()),
                    len(self.balls),
                    round(image_cache.default_expiration, 2) if image_cache.age_cap_is_enabled() else 'off',
                    image_cache.max_memory / megabyte if image_cache.memory_cap_is_enabled() else 'off',
                    image_cache.count(),
                    int(round(image_cache.size() / 1048576)),
                    image_cache.hit_ratio(),
                    image_cache.mem_purged,
                    image_cache.old_purged))

    def draw(self):
        """typical draw routine"""
        screen = pygame.display.get_surface()
        screen.fill((0, 0, 0))
        for ball in self.balls:
            screen.blit(ball.image, ball.rect)
        pygame.display.flip()

    def make_balls(self, num):
        for i in range(num):
            self.balls.append(Ball())

    def remove_balls(self, num):
        for n in range(num):
            i = random.randrange(0, len(self.balls))
            del self.balls[i]


if __name__ == '__main__':
    pygame.init()
    pygame.display.set_mode((800, 600))
    load_base_images()
    game = Game()
    game.run()
