import json

import pygame

import sqlnub


__version__ = '1.0.0'
__vernum__ = __version__.split('.')


DBINFO_COLUMNS = (
    ('name', 'text'),
    ('version', 'text'),
)

IMAGES_COLUMNS = (
    ('name', 'text'),
    ('transforms', 'text'),
    ('width', 'int'),
    ('height', 'int'),
    ('data', 'text'),
    ('primary key', "('name', 'transforms')"),
)


class ImageCacheDB(object):
    """a database for storing and retrieving images and their transformations

    ImageCacheDB features:
        * automated database creation, modestly customizable
        * optional: additional table creation, modestly customizable
        * optional: multiple tables
        * get one or more images
        * save one or more images
        * delete one or more images
        * arbitrary transform descriptors: any json-compatible type
        * saving the contents of an imagecache.ImageCache
        * loading the database into an imagecache.ImageCache
        * auto- and manual commit
        * various info queries

    Use of imagecache.ImageCache is only required for the save() and load()
    methods. save() and load() are conveniences. One can work easily with
    images and database cursors without an ImageCache.

    Updating the records in a database is not supported. If it is needed,
    please DIY.
    """

    def __init__(self, filename, table_name='images', dbinfo_columns=None, images_columns=None,
                 auto_commit=True):
        """open a sqlite3 database for storing and saving images

        :param filename: string; database file name
        :param table_name: optional string; the default table name methods
        should use
        :param dbinfo_columns: sequence of tuple; specifies columns created in
        dbinfo table; default is DBINFO_COLUMNS
        :param images_columns: sequence of tuple; specifies columns create in
        images table(s); default is IMAGES_COLUMNS
        :param auto_commit: bool; if True, each method that modifies the table
        will auto-commit; default is True
        :return: ImageCacheDB

        By default, one images table is created. It is named 'images' unless
        a custom name is identified in the constructor. More than one table can
        be created. If using more than one table, one must supply the table
        name to each method if the table differs from the one identified in the
        constructor. The default table name can be modified at any time by
        assigning a new string to the table_name attribute.

        If auto_commit is False, it is the user's responsibility to commit
        transactions. The need to commit can be detected when the commit_needed
        attribute is True. Auto-commit can be enabled/disabled at any time by
        assigning True or False to the auto_commit attribute.

        A note regarding pygame surfaces. Colorkey is encoded in the surface
        data emitted by Surface.to_string(). Alpha is not so encoded. Thus,
        alpha value is not restored when the image is decoded from the database.
        This applies only to surface alpha, not srcalpha.

        dbinfo_columns and images_columns are merely a provision to guide the
        initial database, and manual creation of additional tables. Additional
        tables can use custom columns by providing column info directly to the
        create_image_table method. However, the use of extra or different
        columns will require a subclass to override relevant methods, direct
        use of sqlite3 or sqlnub, or direct use of a cursor.
        """
        if dbinfo_columns is None:
            dbinfo_columns = DBINFO_COLUMNS
        if images_columns is None:
            images_columns = IMAGES_COLUMNS
        self.image_cache_db = sqlnub.SqlNub(filename)
        self.filename = filename
        self.table_name = table_name
        self.dbinfo_columns = dbinfo_columns
        self.images_columns = images_columns
        self.auto_commit = auto_commit
        self.commit_needed = False
        self._init_database()

    def get(self, name, transforms, table_name=None):
        """get a single image

        :param name: string; name of image
        :param transforms: immutable; anything describing the transformation used
        :param table_name: optional string; name of table to query; default
        taken from __init__
        :return: pygame.Surface
        """
        names_transforms = ((name, transforms),)
        images = self.get_many(names_transforms, table_name)
        image = None
        if images:
            image = images[0]
        return image

    def get_many(self, names_transforms, table_name=None):
        """get many images in a batch

        :param names_transforms: sequence of tuple; e.g. [(name1, transforms1), ...]
        :param table_name: optional string; name of table to query; default
        taken from __init__
        :return: sequence of pygame.Surface
        """
        cursor = self.select_many(names_transforms, table_name)
        images = []
        for row in cursor.fetchall():
            data = row['data']
            size = row['width'], row['height']
            image = pygame.image.fromstring(str(data), size, 'RGBA')
            images.append(image)
        return images

    def get_by_name(self, name, table_name=None):
        """get many images by name

        :param name: string; name of image
        :param table_name: optional string; name of table to query; default
        taken from __init__
        :return: sequence of pygame.Surface
        """
        cursor = self.select_by_name(name, table_name)
        images = []
        for row in cursor.fetchall():
            data = row['data']
            size = row['width'], row['height']
            image = pygame.image.fromstring(str(data), size, 'RGBA')
            images.append(image)
        return images

    def insert(self, name, transforms, image, table_name=None):
        """insert a single record

        :param name: string; name of image
        :param transforms: immutable; anything describing the transformation used
        :param image: pygame.Surface; the image to store
        :param table_name: optional string; name of table to modify; default
        taken from __init__
        :return: None
        """
        names_transforms_images = ((name, transforms, image),)
        self.insert_many(names_transforms_images, table_name)

    def insert_many(self, names_transforms_images, table_name=None):
        """insert many records in a batch

        :param names_transforms_images: sequence of tuple; e.g.
        [(name1, transforms1, image1), ...]
        :param table_name: optional string; name of table to modify; default
        taken from __init__
        :return: None
        """
        params = []
        for name, transforms, image in names_transforms_images:
            data = pygame.image.tostring(image, 'RGBA')
            width, height = image.get_size()
            params.append((name, json.dumps(transforms), width, height, buffer(data)))
        if table_name is None:
            table_name = self.table_name
        sql = """
            insert into {}
            (name, transforms, width, height, data)
            values (?, ?, ?, ?, ?)
            """.format(table_name)
        self.image_cache_db.executemany(sql, params)
        self.commit_needed = True
        if self.auto_commit:
            self.commit()

    def select(self, name, transforms, table_name=None):
        """select a single record

        :param name: string; name of image
        :param transforms: immutable; anything describing the transformation used
        :param table_name: optional string; name of table to query; default
        taken from __init__
        :return: sqlite3.Cursor
        """
        names_transforms = ((name, transforms),)
        return self.select_many(names_transforms, table_name)

    def select_many(self, names_transforms, table_name=None):
        """select many records in a batch

        :param names_transforms: sequence of tuple; e.g. [(name1, transforms1), ...]
        :param table_name: optional string; name of table to query; default
        taken from __init__
        :return: sqlite3.Cursor
        """
        params = []
        conditions = []
        for name, transforms in names_transforms:
            params.extend((name, json.dumps(transforms)))
            conditions.append('(name=? and transforms=?)')
        if table_name is None:
            table_name = self.table_name
        sql = """
            select * from {}
            where {}
            """.format(table_name, ' or '.join(conditions))
        return self.image_cache_db.execute(sql, params)

    def select_by_name(self, name, table_name=None):
        """select records by name

        :param name: string; name of image
        :param table_name: optional string; name of table to query; default
        taken from __init__
        :return: sqlite3.Cursor
        """
        if table_name is None:
            table_name = self.table_name
        sql = """
            select * from {}
            where name=?
            """.format(table_name)
        return self.image_cache_db.execute(sql, (name,))

    def delete(self, name, transforms, table_name=None):
        """delete a single record

        :param name: string; name of image
        :param transforms: immutable; anything describing the transformation used
        :param table_name: optional string; name of table to modify; default
        taken from __init__
        :return: sqlite3.Cursor
        """
        names_transforms = ((name, transforms),)
        self.delete_many(names_transforms, table_name)

    def delete_many(self, names_transforms, table_name=None):
        """delete many records in a batch

        :param names_transforms: sequence of tuple; e.g. [(name1, transforms1), ...]
        :param table_name: optional string; name of table to modify; default
        taken from __init__
        :return: sqlite3.Cursor
        """
        params = []
        for name, transforms in names_transforms:
            params.append((name, json.dumps(transforms)))
        if table_name is None:
            table_name = self.table_name
        sql = """
            delete from {}
            where name=? and transforms=?
            """.format(table_name)
        self.image_cache_db.executemany(sql, params)
        self.commit_needed = True
        if self.auto_commit:
            self.commit()

    def delete_by_name(self, name, table_name=None):
        """delete records by name

        :param name: string; name of image
        :param table_name: optional string; name of table to modify; default
        taken from __init__
        :return: sqlite3.Cursor
        """
        if table_name is None:
            table_name = self.table_name
        sql = """
            delete from {}
            where name=?
            """.format(table_name)
        self.image_cache_db.execute(sql, (name,))
        self.commit_needed = True
        if self.auto_commit:
            self.commit()

    def commit(self):
        """commit transactions

        Sets internal state self.commit_needed to False.

        :return: None
        """
        self.image_cache_db.commit()
        self.commit_needed = False

    def count(self, table_name=None):
        """get the number of records

        :param table_name: optional string; name of table to query; default
        taken from __init__
        :return: int; number of records
        """
        if table_name is None:
            table_name = self.table_name
        cursor = self.image_cache_db.execute('select count(*) from {}'.format(table_name))
        row = cursor.fetchone()
        return row[0]

    def cursor(self):
        """get a database cursor

        :return: sqlite3.Cursor
        """
        return self.image_cache_db.cursor()

    def get_names(self, table_name=None):
        """get a list of the names in a table

        :param table_name: optional string; name of table to query; default
        taken from __init__
        :return: tuple of names
        """
        if table_name is None:
            table_name = self.table_name
        sql = """
            select distinct name from {}
            """.format(table_name)
        cursor = self.image_cache_db.execute(sql)
        return tuple([row['name'] for row in cursor.fetchall()])

    def load(self, image_cache, table_name=None):
        """load an entire table into an image cache

        :param image_cache: imagecache.ImageCache
        :param table_name: optional string; name of table to insert; default
        taken from __init__
        :return: imagecache.ImageCache
        """
        if table_name is None:
            table_name = self.table_name
        cursor = self.image_cache_db.execute('select * from {}'.format(table_name))
        for row in cursor.fetchall():
            name = row['name']
            transforms = json.loads(row['transforms'])
            data = row['data']
            size = row['width'], row['height']
            image = pygame.image.fromstring(str(data), size, 'RGBA')
            image_cache.add(name, transforms, image, -1)
        return image_cache

    def save(self, image_cache, table_name=None):
        """save an entire image cache into a table

        :param image_cache: imagecache.ImageCache
        :param table_name: optional string; name of table to query; default
        taken from __init__
        :return: sqlite3.Cursor
        """
        if table_name is None:
            table_name = self.table_name
        sql = """
            insert into {}
            (name, transforms, width, height, data)
            values (?, ?, ?, ?, ?)
            """.format(table_name)
        params = []
        for name, transforms, image in image_cache.get_entries():
            width, height = image.get_size()
            data = pygame.image.tostring(image, 'RGBA')
            params.append((name, json.dumps(transforms), width, height, buffer(data)))
        cursor = self.image_cache_db.executemany(sql, params)
        self.commit()
        return cursor

    def create_images_table(self, table_name, images_columns=None):
        """create a table for storing images

        :param table_name: string; name of table to create
        :param images_columns: optional sequence; e.g. [(name1, datatype1), ...]; if None,
        self.images_columns from the constructor is used (see __init__)
        :return: None
        """
        if images_columns is None:
            images_columns = self.images_columns
        self.image_cache_db.create_table(table_name, images_columns)
        # self.image_cache_db.execute(
        #     """
        #     create index {}index on {} (name, transforms)
        #     """.format(table_name, table_name))
        self.commit()

    def _init_database(self):
        """initialize a new database

        The filename is taken from __init__.

        The dbinfo table takes columns from __init__.

        The images table (customizable name, see __init__) takes columns from __init__.

        :return: None
        """
        if self.table_name in self.image_cache_db.get_table_names():
            return

        self.image_cache_db.create_table('dbinfo', DBINFO_COLUMNS)
        self.image_cache_db.execute(
            'insert into dbinfo values (?, ?)', (self.filename, __version__))

        self.create_images_table(self.table_name)
